@@ -22,7 +22,8 @@ along with ProtonVPN.  If not, see <https://www.gnu.org/licenses/>.
 from __future__ import annotations
 
 import asyncio
-import copy
+from copy import deepcopy
+import os
 import threading
 from typing import Optional, runtime_checkable, Protocol
 
@@ -33,6 +34,7 @@ from proton.vpn.connection.persistence import ConnectionPersistence
 from proton.vpn.core.refresher import VPNDataRefresher
 from proton.vpn.core.session_holder import SessionHolder
 from proton.vpn.core.settings import SettingsPersistence
+from proton.vpn.core.settings.split_tunneling import SplitTunneling as SplitTunnelingSetting
 from proton.vpn.killswitch.interface import KillSwitch
 
 from proton.vpn import logging
@@ -48,6 +50,8 @@ from proton.vpn.session.dataclasses import VPNLocation
 from proton.vpn.session.servers import LogicalServer, ServerFeatureEnum
 from proton.vpn.core.usage import UsageReporting
 from proton.vpn.connection.exceptions import FeatureSyntaxError, FeatureError
+from proton.vpn.split_tunneling.interface import SplitTunneling
+from proton.vpn.core.cache_handlers import PortForwardFileHandler
 
 logger = logging.getLogger(__name__)
 
@@ -85,11 +89,13 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
         """
         Builds a VPN connector instance and initializes it.
         """
+        split_tunneling = await SplitTunneling.get(os.getuid())
         connector = VPNConnector(
             session_holder,
             settings_persistence,
             kill_switch=kill_switch,
             usage_reporting=usage_reporting,
+            split_tunneling=split_tunneling
         )
         await connector.initialize_state()
         return connector
@@ -99,35 +105,36 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
             session_holder: SessionHolder,
             settings_persistence: SettingsPersistence,
             usage_reporting: UsageReporting,
-            connection_persistence: ConnectionPersistence = None,
-            state: states.State = None,
-            kill_switch: KillSwitch = None,
-            publisher: Publisher = None,
+            connection_persistence: Optional[ConnectionPersistence] = None,
+            state: Optional[states.State] = None,
+            kill_switch: Optional[KillSwitch] = None,
+            split_tunneling: Optional[SplitTunneling] = None,
+            publisher: Optional[Publisher] = None,
+            port_forward_file_handler: PortForwardFileHandler = None,
     ):
         self._session_holder = session_holder
         self._settings_persistence = settings_persistence
         self._connection_persistence = connection_persistence or ConnectionPersistence()
         self._current_state = state
         self._kill_switch = kill_switch
+        self._split_tunneling = split_tunneling
         self._publisher = publisher or Publisher()
         self._lock = asyncio.Lock()
         self._background_tasks = set()
         self._usage_reporting = usage_reporting
+        self._port_forward_file_handler = port_forward_file_handler or PortForwardFileHandler()
 
-        self._publisher.register(self._on_state_change)
+        self._publisher.register(self._on_state_change_update_location)
+        self._publisher.register(self._port_forward_file_handler.on_state_change_update_port)
 
-    def _filter_features(self, input_settings: Settings, user_tier: int = None) -> Settings:
-        if not user_tier:
-            user_tier = self._session_holder.user_tier or 0
+    @property
+    def is_split_tunneling_available(self) -> bool:
+        """Returns if split tunneling is available or not."""
+        split_tunneling_ff = self._get_split_tunneling_feature_flag()
+        return bool(self._split_tunneling) and split_tunneling_ff
 
-        settings = copy.deepcopy(input_settings)
-
-        if self._is_free_tier(user_tier):
-            # Our servers do not allow setting connection features on the free
-            # tier, not even the defaults.
-            settings.features = None
-
-        return settings
+    def _get_split_tunneling_feature_flag(self):
+        return self._session_holder.session.feature_flags.get("DisplaySplitTunneling")
 
     async def get_settings(self) -> Settings:
         """Returns the user's settings."""
@@ -135,21 +142,23 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
         user_tier = self._session_holder.user_tier or 0
         loop = asyncio.get_running_loop()
         settings = await loop.run_in_executor(
-            None, self._settings_persistence.get, user_tier
+            None, self._settings_persistence.get,
+            user_tier,
+            self._session_holder.session.feature_flags
         )
 
-        return self._filter_features(settings, user_tier)
+        return settings
 
     @property
     def credentials(self) -> Optional[VPNCredentials]:
         """Returns the user's credentials."""
         return self._session_holder.vpn_credentials
 
-    def _set_ks_setting(self, settings: Settings):
-        StateContext.kill_switch_setting = KillSwitchSetting(settings.killswitch)
+    def _set_ks_setting(self, ks_setting: KillSwitchSetting, protocol: str):
+        StateContext.kill_switch_setting = ks_setting
 
         if isinstance(self.current_state, states.Disconnected):
-            self._set_ks_impl(settings)
+            self._set_ks_impl(protocol)
 
     async def update_credentials(self):
         """
@@ -167,13 +176,19 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
         Sets the settings to be applied when establishing the next connection and
         applies them to the current connection whenever that's possible.
         """
-        self._set_ks_setting(settings)
-        await self._apply_kill_switch_setting(KillSwitchSetting(settings.killswitch))
-        if self.current_connection:
+        ks_setting = KillSwitchSetting(settings.killswitch)
+        protocol = settings.protocol
+        self._set_ks_setting(ks_setting, protocol)
+        await self._apply_kill_switch_setting(ks_setting)
 
-            await self.current_connection.update_settings(
-                self._filter_features(settings)
-            )
+        if self.current_connection:
+            await self.current_connection.update_settings(settings)
+
+        st_setting = settings.features.split_tunneling
+        self._set_split_tunneling_setting(st_setting)
+        # nosemgrep: python.lang.maintainability.is-function-without-parentheses.is-function-without-parentheses  # pylint: disable=line-too-long  # noqa: E501
+        if self.is_split_tunneling_available and self.is_connected:
+            await self._apply_split_tunneling_settings(st_setting, ks_setting)
 
     async def _apply_kill_switch_setting(self, kill_switch_setting: KillSwitchSetting):
         """Enables/disables the kill switch depending on the setting value."""
@@ -204,6 +219,27 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
         else:
             raise RuntimeError(f"Unexpected kill switch setting: {kill_switch_setting}")
 
+    async def _apply_split_tunneling_settings(
+            self, st_settings: SplitTunnelingSetting, ks_setting: KillSwitchSetting
+    ):
+        if ks_setting != KillSwitchSetting.OFF:
+            logger.warning("Split tunneling is not compatible with the kill switch feature")
+            return
+
+        if not st_settings.enabled:
+            await self._split_tunneling.clear_config()
+        else:
+            await self._split_tunneling.set_config(
+                st_settings.get_config()
+            )
+
+    def get_certificate_based_openvpn(self):
+        """
+        Returns whether the certificate based OpenVPN should be used or not.
+        """
+        feature_flags = self._session_holder.session.feature_flags
+        return feature_flags.get("CertificateBasedOpenVPNWithLocalAgent")
+
     async def _get_current_connection(self) -> Optional[VPNConnection]:
         """
         :return: the current VPN connection or None if there isn't one.
@@ -224,11 +260,15 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
         settings = await self.get_settings()
         for protocol in all_protocols:
             if protocol.cls.protocol == persisted_parameters.protocol:
+
+                use_certificate = self.get_certificate_based_openvpn()
+
                 vpn_connection = protocol.cls(
                     server=persisted_parameters.server,
                     credentials=self.credentials,
                     settings=settings,
-                    connection_id=persisted_parameters.connection_id
+                    connection_id=persisted_parameters.connection_id,
+                    use_certificate=use_certificate
                 )
                 if not isinstance(vpn_connection.initial_state, states.Disconnected):
                     return vpn_connection
@@ -237,10 +277,15 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
 
     async def _get_initial_state(self):
         """Determines the initial state of the state machine."""
-        current_connection = await self._get_current_connection()
 
-        if current_connection:
-            return current_connection.initial_state
+        # It's possible that the user is not logged in but that there is
+        # a persisted connection, in this case we need to ignore the persisted
+        # connection and return the disconnected state.
+        if self._session_holder.session.logged_in:
+            current_connection = await self._get_current_connection()
+
+            if current_connection:
+                return current_connection.initial_state
 
         return states.Disconnected(
             StateContext(event=events.Initialized(events.EventContext(connection=None)))
@@ -252,7 +297,10 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
 
         settings = await self.get_settings()
         StateContext.kill_switch_setting = KillSwitchSetting(settings.killswitch)
-        self._set_ks_impl(settings)
+        self._set_ks_impl(settings.protocol)
+
+        self._set_split_tunneling_setting(settings.features.split_tunneling)
+        self._set_split_tunneling_impl()
 
         connection = state.context.connection
         if connection:
@@ -356,17 +404,13 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
 
         # Sets the settings to be applied when establishing the next connection.
         settings = await self.get_settings()
-        self._set_ks_setting(settings)
+        # FIXME: this adds a big delay before creating the connection  # pylint: disable=fixme
+        self._set_ks_setting(KillSwitchSetting(settings.killswitch), settings.protocol)
+        self._set_split_tunneling_setting(settings.features.split_tunneling)
 
         protocol = protocol or settings.protocol
 
-        # If IPv6 FF is disabled then the feature should not be toggled client side and
-        # should be disabled.
-        if not self._can_ipv6_be_toggled_client_side(settings):
-            settings.ipv6 = False
-
-        feature_flags = self._session_holder.session.feature_flags
-        use_certificate = feature_flags.get("CertificateBasedOpenVPN")
+        use_certificate = self.get_certificate_based_openvpn()
 
         logger.info("Using certificate based authentication"
                     f" for openvpn: {use_certificate}")
@@ -467,35 +511,64 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
             # Unregister from connection event updates once the connection ended.
             self._current_state.context.connection.unregister(self._on_connection_event)
 
-        new_event = await self._current_state.run_tasks()
-        self._publisher.notify(new_state)
+        if isinstance(old_state, states.Connected) and isinstance(new_state, states.Connected):
+            # A Connected state can transition to a new Connected state when local agent
+            # sends a new connected event, e.g. with a new port forwarding port. In this case,
+            # the connection subscribers are notified (to update data on the client UI)
+            # but the tasks associated with the connected state do not need to be run again
+            # since the connection state didn't really change, only its context data did.
+            self._publisher.notify(new_state)
+            return None
+
+        if self._current_state.notify_early:
+            self._publisher.notify(new_state)
+            new_event = await self._current_state.run_tasks()
+        else:
+            new_event = await self._current_state.run_tasks()
+            self._publisher.notify(new_state)
 
         if (
             not self._current_state.context.reconnection
             and isinstance(self._current_state, states.Disconnected)
         ):
-            self._set_ks_impl(await self.get_settings())
+            self._set_ks_impl((await self.get_settings()).protocol)
 
         return new_event
 
-    def _on_state_change(self, state: states.State):
+    def _on_state_change_update_location(self, state: states.State):
         """Updates the user location when the connection is established."""
-        if not isinstance(state, states.Connected):
-            return
-
-        connection_details = state.context.event.context.connection_details
-        if not connection_details or not connection_details.device_ip:
+        connection_details = self._get_connection_details_from_state(state)
+        if not connection_details:
             return
 
         current_location = self._session_holder.session.vpn_account.location
-        vpnlocation = VPNLocation(
+
+        self._session_holder.session.set_location(
+            self._create_new_vpn_location(connection_details, current_location)
+        )
+
+    def _get_connection_details_from_state(
+            self, state: states.State
+    ) -> Optional[events.ConnectionDetails]:
+        if not isinstance(state, states.Connected):
+            return None
+
+        connection_details = state.context.event.context.connection_details
+        if not connection_details or not connection_details.device_ip:
+            return None
+
+        return connection_details
+
+    def _create_new_vpn_location(self, connection_details, current_location) -> VPNLocation:
+        return VPNLocation(
             IP=connection_details.device_ip,
             Country=connection_details.device_country,
-            ISP=current_location.ISP
+            ISP=current_location.ISP,
+            Long=current_location.Long,
+            Lat=current_location.Lat
         )
-        self._session_holder.session.set_location(vpnlocation)
 
-    def _set_ks_impl(self, settings: Settings):
+    def _set_ks_impl(self, protocol: str):
         """
         By using this specific method we're leaking implementation details.
 
@@ -505,16 +578,31 @@ class VPNConnector:  # pylint: disable=too-many-instance-attributes
         we only do this when we are in `Disconnected` state, to ensure
         that the environment is clean and we don't leave any residuals on a users machine.
         """
-        protocol = settings.protocol
         kill_switch_backend = KillSwitch.get(protocol=protocol)
         StateContext.kill_switch = self._kill_switch or kill_switch_backend()
 
-    def _is_free_tier(self, user_tier: int) -> bool:
-        return user_tier == 0
+    def _set_split_tunneling_setting(self, st_setting: SplitTunnelingSetting):
+        split_tunneling_ff = self._get_split_tunneling_feature_flag()
 
-    def _can_ipv6_be_toggled_client_side(self, settings: Settings) -> bool:
-        return settings.ipv6 and\
-            self._session_holder.session.feature_flags.get("IPv6Support")
+        st_setting = deepcopy(st_setting)
+        st_setting.enabled = (
+            self._split_tunneling
+            and st_setting.enabled
+            and not self._is_free_tier()
+            and split_tunneling_ff
+        )
+
+        StateContext.split_tunneling_setting = st_setting
+
+    def _set_split_tunneling_impl(self):
+        StateContext.split_tunneling = self._split_tunneling
+
+    def _get_user_tier(self) -> int:
+        # Default to free tier if session is not loaded yet
+        return self._session_holder.user_tier or 0
+
+    def _is_free_tier(self) -> bool:
+        return self._get_user_tier() == 0
 
     def subscribe_to_certificate_updates(self, refresher: VPNDataRefresher):
         """Subscribes to certificate updates."""
